#include <stdlib.h>
#include <stdio.h>

#include <GL/gl.h>
#include <GL/glut.h>
#include <math.h>

#define IX(i,j) ((i)+(N+2)*(j))


/* Declare external functions from fluidsolver.c */
extern void dens_step ( float * x, float * x0, float * u, float * v );
extern void vel_step ( float * u, float * v, float * u0, float * v0, float * theta, float * dens, float* curl);
extern void setup ( int N, float dt, float visc, float diff );

/* global variables */
static int N;
static float dt, diff, visc;
static float force, source;
static float * u, * v, * u_prev, * v_prev;
static float * dens, * dens_prev, *curl;
static float * theta, *theta_prev;  // We don't need this since the density represents the same thing

/* global variables for UI */
static int win_x, win_y, win_id;
static int mouse_down[3];
static char titleText[50];
static int omx, omy, mx, my;
int now, lastFrameTime, elapsedMilliseconds;
int frames;

/**************************************************************
 * Function: free_data
 * Definition: Used to free data
 *************************************************************/
static void free_data ( void )
{
	if ( u ) free ( u );
	if ( v ) free ( v );
	if ( u_prev ) free ( u_prev );
	if ( v_prev ) free ( v_prev );
	if ( dens ) free ( dens );
	if ( theta ) free ( theta );
	if ( dens_prev ) free ( dens_prev );
	if ( curl ) free ( curl );
	if ( theta_prev ) free ( theta_prev );
}

/*************************************************************
 * Function: clear_data
 * Definition: to clear data (set everything to zero) ie. clear screen
 *************************************************************/
static void clear_data ( void )
{
	int i, size=(N+2)*(N+2);

	for ( i=0 ; i<size ; i++ ) {
		u[i] = v[i] = u_prev[i] = v_prev[i] = dens[i] = dens_prev[i] = theta[i] = dens_prev[i] = 0.0f;
	}
}

/* Allocate the needed data */
static int allocate_data ( void )
{
	/* size of grid */
	int size = (N+2)*(N+2);

	u			= (float *) malloc ( size*sizeof(float) );
	v			= (float *) malloc ( size*sizeof(float) );
	u_prev		= (float *) malloc ( size*sizeof(float) );
	v_prev		= (float *) malloc ( size*sizeof(float) );
	dens		= (float *) malloc ( size*sizeof(float) );	
	dens_prev	= (float *) malloc ( size*sizeof(float) );
	theta		= (float *) malloc ( size*sizeof(float) );
	theta_prev 	= (float *) malloc ( size*sizeof(float) );
	curl		= (float *) malloc ( size*sizeof(float) );

	/* If allocation didn't succeed */
	if ( !u || !v || !u_prev || !v_prev || !dens || !dens_prev || !curl )
	{
		fprintf ( stderr, "cannot allocate data\n" );
		return ( 0 );
	}

	return ( 1 );
}

/**************************************************************
 * Function: get_from_UI
 * Definition: to capture input from User Interface
 *************************************************************/
static void get_from_UI ( float * d, float *t, float * u, float * v )
{
	/* 
	 * Create grid for calculations and add 2 extra cells in (x,y) so
	 * we have an empty boundary (frame)
	 */
	int i, j, size = (N+2)*(N+2);

	/* clear matrix */
	for ( i=0 ; i<size ; i++ ) 
	{
		u[i] = v[i] = d[i] = t[i] = 0.0f;
	}

	/* If no button is pushed, do nothing */
	if ( !mouse_down[0] && !mouse_down[2] )
	{
		return;
	}

	/* Calculate coord of mouseclick */
	i = (int)((       mx /(float)win_x)*N+1);
	j = (int)(((win_y-my)/(float)win_y)*N+1);

	if ( i<1 || i>N || j<1 || j>N )
	{
		fprintf(stdout, "Can't deploy smoke outside borders\n");
		return;
	}

	/* If left button is pressed, apply force to area */
	if ( mouse_down[0] )
	{
		u[IX(i,j)] = force * (mx-omx);
		v[IX(i,j)] = force * (omy-my);
	}

	/* If right mouse button is pressed, deploy smoke */
	if ( mouse_down[2] )
	{
		d[IX(i,j)] = source;
		t[IX(i,j)] = 373; //deploy heat at 100 C
	}

	/* Save old coordinates */
	omx = mx;
	omy = my;

	return;
}

/*************************************************************
 * Function: draw_density
 * Definition: draws the smoke 
 *************************************************************/
void draw_density ( void ) 
{

	int i, j;
	float h = 1.0f/N, x, y, d00, d01, d10, d11;

	glBegin( GL_QUADS );

	/* Draw screen here! */
	for (i = 0; i < N; i++) {
		x = (i-0.5f)*h; /* calculate the x-pos of the vertices of the quad */
		for (j = 0; j < N; j++) {
			y = (j-0.5f)*h; /* calculate the y-pos of the vertices of the quad */

			/* Get the color of each vertex in the quad */

			d00 = dens[IX(i,j)];
			d01 = dens[IX(i,j+1)];
			d10 = dens[IX(i+1,j)];
			d11 = dens[IX(i+1,j+1)];
/*
			d00 = theta[IX(i,j)];
			d01 = theta[IX(i,j+1)];
			d10 = theta[IX(i+1,j)];
			d11 = theta[IX(i+1,j+1)];
*/
			/* Set the color -> draw the vertex
			 * OpenGL automatically "interpolates" between the vertices */
			glColor3f ( d00, d00, d00 ); glVertex2f ( x, y );
			glColor3f ( d10, d10, d10 ); glVertex2f ( x+h, y );
			glColor3f ( d11, d11, d11 ); glVertex2f ( x+h, y+h );
			glColor3f ( d01, d01, d01 ); glVertex2f ( x, y+h );
		}
	}
	glEnd();
}

/*************************************************************
 * Function: mouse_pressed
 * Definition: Used for capturing mouse clicks 
 *************************************************************/
static void mouse_pressed ( int button, int state, int x, int y )
{

	omx = mx = x;
	omy = my = y;

	mouse_down[button] = state == GLUT_DOWN;

}

/*************************************************************
 * Function: mouse_moved
 * Definition: Captures the difference in mouse movement 
 *************************************************************/
static void mouse_moved ( int x, int y)
{

	mx = x;
	my = y;

}

/*************************************************************
 * Function: key_func
 * Definition: Captures key strokes from the keyboard and
 *					behaves thereafter 
 ************************************************************/
static void key_func ( unsigned char key, int x, int y )
{
	switch ( key )
	{
		case 'c':
		case 'C':
			clear_data ();
			break;

		case 'q':
		case 'Q':
			free_data ();
			exit ( 0 );
			break;
	}
}

/*************************************************************
 * Function: reshape_func
 * Definition: Used to handle resizing of the window
 *************************************************************/
static void reshape_func ( int width, int height )
{
	glutSetWindow ( win_id );
	glutReshapeWindow ( width, height );

	win_x = width;
	win_y = height;
}

/*************************************************************
 * Function: idle_func
 * Definition: The main glut function, is repeated
 *					round-n-round-n-round 
 *************************************************************/
static void idle_func ( void )
{
	get_from_UI ( dens_prev, theta_prev, u_prev, v_prev ); /* get input from UI */
	vel_step ( u, v, u_prev, v_prev, theta, dens, curl ); /* perform one velocity step calculation */
	dens_step ( dens, dens_prev, u, v ); /* perform one density step calculation */
	dens_step ( theta, theta_prev, u, v ); /* perform one heat step calculation */

	/* redraw window etc */
	glutSetWindow ( win_id );
	glutPostRedisplay ();
}

/*************************************************************
 * Function: calc_framerate
 * Definition: Calculates the framerate 
 ************************************************************/
void update_titlebar ( void )
{

	now = glutGet(GLUT_ELAPSED_TIME);
	elapsedMilliseconds = now - lastFrameTime;

	if (elapsedMilliseconds > 1000 || frames == 0)
	{
		// Update fps
		double fps = ((double)frames/(double)elapsedMilliseconds)*1000.0;
		sprintf(titleText, "Smokesimulation - %0.1f FPS", fps);
		glutSetWindowTitle(titleText);
		lastFrameTime = now;
		frames = 0;
	}
	++frames;

}

/*************************************************************
 * Function: display_func
 * Definition: Updates the screen and the titlebar 
 *************************************************************/
static void display_func ( void )
{
	/* The drawing function */
	draw_density ();
	update_titlebar();

	/* Swap drawing buffers */
	glutSwapBuffers ();
}

/**************************************************************
 * Function: open_glut_window
 * Definition: open a glut compatible window and set callbacks
 *************************************************************/
static void open_glut_window ( void )
{
	glutInitDisplayMode ( GLUT_RGBA | GLUT_DOUBLE );

	glutInitWindowPosition ( 0, 0 );
	glutInitWindowSize ( win_x, win_y );
	win_id = glutCreateWindow ( titleText );

	glClearColor ( 0.0f, 0.0f, 0.0f, 1.0f );
	glClear ( GL_COLOR_BUFFER_BIT );

	glViewport ( 0, 0, win_x, win_y );
	glMatrixMode ( GL_PROJECTION );
	glLoadIdentity ();
	gluOrtho2D ( 0.0, 1.0, 0.0, 1.0 );
	glClearColor ( 0.0f, 0.0f, 0.0f, 1.0f );
	glClear ( GL_COLOR_BUFFER_BIT );

	glutKeyboardFunc ( key_func );
	glutMouseFunc ( mouse_pressed );
	glutMotionFunc ( mouse_moved );
	glutReshapeFunc ( reshape_func );
	glutIdleFunc ( idle_func );
	glutDisplayFunc ( display_func );
}

/***************************************************
 *	Main function
 * Description: Does what a main functions does best
 **************************************************/
int main (int argc, char **argv)
{

	glutInit ( &argc, argv );

	if ( argc != 1 && argc != 7 ) {
		fprintf ( stderr, "usage : %s N dt diff visc force source\n", argv[0] );
		fprintf ( stderr, "where:\n" );
		fprintf ( stderr, "\t N      : grid resolution\n" );
		fprintf ( stderr, "\t dt     : time step\n" );
		fprintf ( stderr, "\t diff   : diffusion rate of the density\n" );
		fprintf ( stderr, "\t visc   : viscosity of the fluid\n" );
		fprintf ( stderr, "\t force  : scales the mouse movement that generate a force\n" );
		fprintf ( stderr, "\t source : amount of density that will be deposited\n" );
		exit ( 1 );
	}

	if ( argc == 1 )
	{
		N = 120;
		dt = 0.2f;
		diff = 0.0f;
		visc = 0.0f;
		force = 5.0f;
		source = 100.0f;
		fprintf ( stderr, "Using defaults : N=%d dt=%g diff=%g visc=%g force = %g source=%g\n", N, dt, diff, visc, force, source );
	} 
	else
	{
		N = atoi(argv[1]);
		dt = atof(argv[2]);
		diff = atof(argv[3]);
		visc = atof(argv[4]);
		force = atof(argv[5]);
		source = atof(argv[6]);
	}

	printf ( "\n\nHow to use this demo:\n\n" );
	printf ( "\t Add densities with the right mouse button\n" );
	printf ( "\t Add velocities with the left mouse button and dragging the mouse\n" );
	printf ( "\t Quit by pressing the 'ESC' key\n\n" );

	if ( !allocate_data () )
	{
		exit ( 1 );
	}

	clear_data ();

	// Setup the global variables
	setup ( N, dt, visc, diff );

	/* Declare variables */
	win_x = win_y = 512;
	lastFrameTime = glutGet(GLUT_ELAPSED_TIME);

	/* Open and initialize the glut window */
	open_glut_window ();

	/* run main loop */
	glutMainLoop ();

	return 0;
}
