/*******************************************
 * 
 * The implementation of a smoke simulation
 * using a semi-Lagrangian Gauss-Seidel Relaxation
 * solver and Incompressible
 * Euler equations etc...
 * 
 ******************************************/
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#define IX(i,j) ((i)+(N+2)*(j))
#define SWAP(x0,x) {float *tmp=x0;x0=x;x=tmp;}

static int N;
static float dt, diff, visc;

/* global constants */
//const float cp = 1006;					/* Specifik heat capacity for dry air (J/kg) at constant pressure */
//const float cv = 718;					/* Specifik heat capacity for dry air (J/kg) at constant pressure */
//const int NORMAL_PRESSURE = 100;		/* pressure at surface */
//const float K = 0.286;
const float AMB_TEMP = 298;				/* Ambient temperature which equals around 25C */
//const int GAS_CONST = 287;				/* Gas constant for dry air */
//const int GAMMA = 10;					/* the Lapse Rate, ie.the rate of decrease of temperature with altitude */
//const float L = 2.501;					/* Amount of energy it takes to vaporize 1 kg of water */

/*******************************************
 * Function: buoyant_force
 * Description: Calculate the Buoyancy force
 ******************************************/
void buoyant_force ( float* buoyancy, float * theta, float * qh )
{
	float size=(N+2)*(N+2);
	float gravity = 9.82f;
	float scale = 0.001f;
	float thetaV = 0.0f;
	float a = 0.000625f;
	float b = 0.025f;
	float ambient_temp = 0.0f;
	float T;

	int i,j;
	for ( i=0; i<size; ++i)
	{
		//if (theta[i] != 0)
			ambient_temp += qh[i];
//		else
//			ambient_temp = ambient_temp + AMB_TEMP;
	}

	ambient_temp /= size;

	for ( i=0; i<N+2; ++i)
	{
		for ( j=0; j<N+2; ++j)
		{
			// calculate virtual potential temperature based on amount of vapor
			// known as ThetaV
			//float p = 100.0f * pow(( 1.0f - j*10.0f/ambient_temp),(9.82f/(10.0f * 287.0f)));
			//T = ambient_temp * pow((p / 100.0f),0.286f);
			//float q_s = (380.16f/p) * exp((17.67*T)/(T+243.5f));
			//theta[IX(i,j)] = T * pow((100.0f / p),0.286);
			//thetaV = theta[IX(i,j)] * (1.0f + 0.61f * q_s );
			//buoyancy[i] = gravity * (thetaV / ambient_temp - q_s);

			// Stam's smoke density
			buoyancy[IX(i,j)] = a * qh[IX(i,j)] + b * (theta[IX(i,j)] - ambient_temp);

			// "Simple" temperature based upwards convection
			// buoyancy[IX(i,j)] = scale * (theta[IX(i,j)] - ambient_temp);
			//	buoyancy[i] = scale * (theta[i] - ambient_temp);
		}
	}

}

/**
 * Calculate the curl at position (i, j) in the fluid grid.
 * Physically this represents the vortex strength at the
 * cell. Computed as follows: w = (del x U) where U is the
 * velocity vector at (i, j).
 **/
float calculate_curl(float* u, float* v, int i, int j)
{
	float du_dy = (u[IX(i, j + 1)] - u[IX(i, j - 1)]) * 0.5f;
	float dv_dx = (v[IX(i + 1, j)] - v[IX(i - 1, j)]) * 0.5f;

	return du_dy - dv_dx;
}

/*******************************************
 * Function: vorticity_confinement
 * Description: Add a force in the direction
 *				the flow directed.
 ******************************************/
void vorticity_confinement ( float* fcv_u, float* fcv_v, float* curl )
{
	float dw_dx, dw_dy;
	float length;
	float v;

	// Calculate magnitude of curl(u,v) for each cell. (|w|)
	for (int i = 1; i <= N; i++)
	{
		for (int j = 1; j <= N; j++)
		{
			curl[IX(i, j)] = abs(calculate_curl(fcv_u, fcv_v, i, j));
		}
	}

	for (int i = 1; i < N; i++)
	{
		for (int j = 1; j < N; j++)
		{

			// Find derivative of the magnitude (n = del |w|)
			dw_dx = (curl[IX(i + 1, j)] - curl[IX(i - 1, j)]) * 0.5f;
			dw_dy = (curl[IX(i, j + 1)] - curl[IX(i, j - 1)]) * 0.5f;

			// Calculate vector length. (|n|)
			// Add small factor to prevent divide by zeros.
			length = (float) sqrt(dw_dx * dw_dx + dw_dy * dw_dy) + 0.000001f;

			// N = ( n/|n| )
			dw_dx /= length;
			dw_dy /= length;

			v = calculate_curl(fcv_u, fcv_v, i, j);

			// N x w
			fcv_u[IX(i, j)] = dw_dy * -v * (float)N * 10.0f;
			fcv_v[IX(i, j)] = dw_dx *  v * (float)N * 10.0f;
		}
	}

}

/*******************************************
 * Function: add_source
 * Description: Add force
 ******************************************/
void add_source ( float * x, float * s ) {
	int i, size=(N+2)*(N+2);
	for ( i=0 ; i<size ; i++ )
		x[i] += dt*s[i];
}

/*******************************************
 * Function: set_bnd
 * Description: Makes the boundaries (outer cells)
 *					act like walls by recalculating the
 *					velocity of the smoke
 ******************************************/
void set_bnd ( int b, float * x )
{
	int i;
	for ( i=1 ; i<=N ; i++ ) {
		// If the smoke is travelling towards the wall
		// revert its flow in opposite direction
		x[IX(0  ,i)] = b==1 ? -x[IX(1,i)] : x[IX(1,i)];
		x[IX(N+1,i)] = b==1 ? -x[IX(N,i)] : x[IX(N,i)];
		x[IX(i, 0 )] = b==2 ? -x[IX(i,1)] : x[IX(i,1)];
		x[IX(i,N+1)] = b==2 ? -x[IX(i,N)] : x[IX(i,N)];
	}
	// Calculate mean values in the corners of the grid
	x[IX(0 ,0 )] = 0.5*(x[IX(1,0 )]+x[IX(0 ,1)]);
	x[IX(0 ,N+1)] = 0.5*(x[IX(1,N+1)]+x[IX(0 ,N )]);
	x[IX(N+1,0 )] = 0.5*(x[IX(N,0 )]+x[IX(N+1,1)]);
	x[IX(N+1,N+1)] = 0.5*(x[IX(N,N+1)]+x[IX(N+1,N )]);
}

/*******************************************
 * Function: diffuse
 * Description: Calculate the diffusion of the smoke
 ******************************************/
void diffuse ( int b, float * x, float * x0 )
{
	int i, j, k;
	float a=dt*diff*N*N;
	for ( k=0 ; k<20 ; k++ ) {
		for ( i=1 ; i<=N ; i++ ) {
			for ( j=1 ; j<=N ; j++ ) {
				x[IX(i,j)] = (x0[IX(i,j)] + a*(x[IX(i-1,j)]+x[IX(i+1,j)]+
						x[IX(i,j-1)]+x[IX(i,j+1)]))/(1+4*a);
			}
		}
		set_bnd ( b, x );
	}
}

/*************************************************
 * Function: advect
 * Description: The advection function that
 *					solves the differential equations
 *					using a semi-lagrangian solver
 ************************************************/
void advect ( int b, float * d, float * d0, float * u, float * v ) 
{
	int i, j, i0, j0, i1, j1;
	float x, y, s0, t0, s1, t1, dt0;
	dt0 = dt*N;
	for ( i=1 ; i<=N ; i++ ) {
		for ( j=1 ; j<=N ; j++ ) {

			x = i-dt0*u[IX(i,j)];
			y = j-dt0*v[IX(i,j)];
			if (x<0.5) x=0.5; if (x>N+0.5) x=N+ 0.5; i0=(int)x; i1=i0+1;
			if (y<0.5) y=0.5; if (y>N+0.5) y=N+ 0.5; j0=(int)y; j1=j0+1;
			s1 = x-i0; s0 = 1-s1; t1 = y-j0; t0 = 1-t1;
			d[IX(i,j)] = s0*(t0*d0[IX(i0,j0)]+t1*d0[IX(i0,j1)]) + s1*(t0*d0[IX(i1,j0)]+t1*d0[IX(i1,j1)]);
		}
	}
	set_bnd ( b, d );
}

/********************************************************
 * Function: dens_step
 * Description: Perform a density time-step calculation
 *
 *******************************************************/
void dens_step ( float * x, float * x0, float * u, float * v, float diff )
{
	add_source ( x, x0 );
	SWAP ( x0, x ); diffuse ( 0, x, x0 );
	SWAP ( x0, x ); advect ( 0, x, x0, u, v );
}

/*******************************************
 * Function: project
 * Description:
 ******************************************/
void project ( float * u, float * v, float * p, float * div )
{
	int i, j, k;
	float h;
	h = 1.0/N;
	for ( i=1 ; i<=N ; i++ ) {
		for ( j=1 ; j<=N ; j++ ) {
			div[IX(i,j)] = -0.5*h*(u[IX(i+1,j)]-u[IX(i-1,j)]+
				v[IX(i,j+1)]-v[IX(i,j-1)]);
			p[IX(i,j)] = 0;
		}
	}
	set_bnd ( 0, div ); set_bnd ( 0, p );
	for ( k=0 ; k<20 ; k++ ) {
		for ( i=1 ; i<=N ; i++ ) {
			for ( j=1 ; j<=N ; j++ ) {
				p[IX(i,j)] = (div[IX(i,j)]+p[IX(i-1,j)]+p[IX(i+1,j)]+
					p[IX(i,j-1)]+p[IX(i,j+1)])/4;
			}
		}
		set_bnd ( 0, p );
	}
	for ( i=1 ; i<=N ; i++ ) {
		for ( j=1 ; j<=N ; j++ ) {
			u[IX(i,j)] -= 0.5*(p[IX(i+1,j)]-p[IX(i-1,j)])/h;
			v[IX(i,j)] -= 0.5*(p[IX(i,j+1)]-p[IX(i,j-1)])/h;
		}
	}
	set_bnd ( 1, u ); set_bnd ( 2, v );
}

/******************************************************
 * Function: vel_step
 * Description: Calculate a velocity time step
 *****************************************************/
void vel_step ( float * u, float * v, float * u0, float * v0, float * theta, float * dens, float* curl )
{

	// Add velocity input by mouse
	add_source ( u, u0 );
	add_source ( v, v0 );

	// Add vorticity confinement
	vorticity_confinement( u0, v0, curl );
	add_source( u, u0 );
	add_source( v, v0 );

	// Add buoyancy force
	buoyant_force ( v0, theta, dens );
	add_source ( v, v0 );

	// First swap to save old variables as u0, v0 respectively
	SWAP ( u0, u );
	diffuse ( 1, u, u0 ); //Calculate diffusion in u
	SWAP ( v0, v );
	diffuse ( 2, v, v0 ); //Calculate diffusion in v
	project ( u, v, u0, v0 ); //Helmut hodge decomposition
	SWAP ( u0, u );
	SWAP ( v0, v );
	advect ( 1, u, u0, u0, v0 );
	advect ( 2, v, v0, u0, v0 );
	project ( u, v, u0, v0 );

	for (int i = 0; i<N*N;++i)
	{
		v0[i] = 0; u0[i] = 0;
	}

}

void setup ( int aN, float adt, float avisc, float adiff )
{
	N = aN;
	dt = adt;
	visc = avisc;
	diff = adiff;
}
